/************************************************************************
*                                                                       *
*    Copyright 2000 LSI Logic Corporation.  All rights reserved.        *
*                                                                       *
*   This file is confidential and a trade secret of LSI Logic           * 
*   Corporation. The receipt of or possession of this file does not     *
*   convey any rights to reproduce or disclose its contents or to       *
*   manufacture, use, or sell anything is may describe, in whole, or in *
*   part, without the specific written consent of LSI Logic Corporation.*
*                                                                       *
************************************************************************/


/*+++HDR
 *           Name:  mpt_dv.c
 *          Title:  MPI Domain Validataion Source
 *       Revision:  0.01
 *        Modtime:  15 August 2000
 *     Programmer:  LSI Logic
 *  Creation Date:
 *
 *  Version History
 *  ---------------
 *
 *    Date    Who?  Description
 *  --------  ----  -------------------------------------------------------
 *
 *-------------------------------------------------------------------------
 *
 */

#include <sys/cmn_err.h>
#include <sys/param.h>
#include <sys/scsi.h>

#include "mpt.h"

#include "typedefs.h"
#include "mptdbg.h"

#include "incl/mpi_type.h"
#include "incl/mpi.h"
#include "incl/mpi_ioc.h"
#include "incl/mpi_init.h"
#include "incl/mpi_cnfg.h"
#include "incl/mpi_raid.h"
#include "devinfo.h"
#include "sim.h"

/****************************************************************************/
/*      Internal Function Prototypes                                        */
/****************************************************************************/


/*
 * DV only routines
 */
void mpt_LUNComplete (char flag, char *msg);
void mpt_DeviceDiscovery (pDev_t dp);
void mpt_DVConfigHeader (void);
void mpt_DVConfigHRCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVReadPage (void);
void mpt_DVReadPageCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVWritePage (void);
void mpt_DVWritePageCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVWritePage4 (void);
void mpt_DVQuiesceEnable (uchar Action );
void mpt_DVQuiesceCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr );
void mpt_DVEnableCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr );
void mpt_DVParms ( uchar flag);
void mpt_DVIoCmd (void);
void mpt_DVTURCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVInqCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVInqLunCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVReadBufferCapCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVReserveCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVWriteBufferCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVReadBufferCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVReleaseCompleted ( pHBA_t mhp, TableStruct_t *MFtabPtr );
void mpt_DVFillBuffer ( uchar width );
void  mpt_setDevParameters ( uchar *from);

/****************************************************************************/
/*           External Function Prototypes                                   */
/****************************************************************************/

extern void  mpt_RemoveDevice( pDev_t dp, uchar lun );
extern void  mpt_StartNewCommand( pHBA_t mhp, TableStruct_t *MFtabPtr );
extern void  mpt_StartQueuedCommands( pHBA_t mhp);

extern TableStruct_t *mpt_GetFreeMF (pHBA_t mhp);
extern void  mpt_FreeCmdMFs ( pHBA_t mhp, ubit32 MFindex);

extern void  mpt_PostRFtoIOC (pHBA_t mhp, ubit32 physAddr);
extern void  mpt_ReadConfigPage (pHBA_t mhp, uchar pageNumber, uchar pageType, uint pageAddress);

#ifdef _HPUX_SOURCE
extern void  *mpt_MemAllocPhys( ubit32 size, ubit32 *physaddr, struct isc_table_type *isc );
#else
extern void  *mpt_MemAllocPhys( ubit32 size, ubit32 *physaddr );
#endif

/****************************************************************************/
/*      External Variables                                                  */
/****************************************************************************/

extern  void     *mpt_HBAs[];  
extern  uchar    mpt_LAST_LUN;
extern  uchar    mpt_DoDomainValidation;
extern  uchar    mpt_Debug_Verbose_Level;


/****************************************************************************/
/*      Local Defines                                                       */
/****************************************************************************/

#define MPT_6BYTE_CMD        6
#define MPT_10BYTE_CMD      10
#define MPT_RETRY_MAX        2
/*
 * DVparms function call defines.
 */
#define MPT_DV_SET_MIN           0x01
#define MPT_DV_SET_MAX           0x02
#define MPT_DV_SET_FALLBACK      0x04
#define MPT_DV_GET_NVRAM_PARMS   0x08

/*
 * DVpass define values
 */
#define MPT_DVPASS_MASK          0x0F
#define MPT_DVPASS_ECHO_BUFFER   0x08
#define MPT_DVPASS_DATA_BUFFER   0x00
#define MPT_DVPASS_CAN_RETRY     0x40

SCSIDevicePage1_t *DevicePageBufPtr;
ubit32 DevicePagePhysAddr;

uchar DVpass;    /* Pass Number for the tests. */
uchar DVBufferPass;

#ifdef MPT_DEBUG
void mpt_dbg_scan()
{
   cmn_err(CE_NOTE, "Debug Stop:");
}
#endif
/****************************************************************************
 *
 * Domain Validation Specific Functionality
 *
 ****************************************************************************/

uchar *FirstDVBufPtr = NULL;
uchar *SecondDVBufPtr;
ubit32 FirstPhysAddr;
ubit32 SecondPhysAddr;
ubit32 DVBufferCapacity; 

ConfigPageHeader_t DVDevicePageHeader;

uchar DVpattern; /* Pattern Number for the w/r compare tests */
uchar DVcmd;     /* Current DV command being issued */
uchar DVpath;    /* path number for DV command */
uchar DVbus;     /* bus number for DV command */
uchar DVid;      /* id number for DV command */
uchar DVlun;     /* lun number for DV command */
uchar DVPhysDiskNum;      /* IM PhysDiskNum */
uchar DVflags;    
ubit32 DVInqLength;
TableStruct_t *DVMFtabPtr;
ubit32  DVMFindex;
pHBA_t DVmhp;
pDev_t DVdp;
ubit32 devNvram;
#ifdef _HPUX_SOURCE
    struct isc_table_type *DVisc;
#endif  
char  flags;

#define MPT_DV_INQUIRY_LENGTH   56
#define MPT_DV_READ_BUFFER_CAP_LENGTH  4

/*
 * DVcmd define values
 */
#define MPT_DV_TUR                       0x00
#define MPT_DV_INQUIRY                   0x12
#define MPT_DV_START                     0x1B
#define MPT_DV_WRITE_BUFFER              0x3B
#define MPT_DV_READ_BUFFER               0x3C
#define MPT_DV_READ_BUFFER_CAPACITY      (0x3C | 0x80)
#define MPT_DV_RELEASE                   0x17
#define MPT_DV_RELEASE_THEN_RESERVE      (0x17 | 0x80)
#define MPT_DV_RESERVE                   0x16

#define MPT_DV_SHOW_MSG     0x80  /* Or'd in with DV status */
#define MPT_DV_SUCCESS      0x01
#define MPT_DV_FAIL         0x02

/*
 * DVflags defines
 */
#define MPT_DV_DEVICE_RESERVED 1
#define MPT_DV_DEVICE_QUIESCED 2
#define MPT_DV_INQ_DONE        4
#define MPT_DV_DONE            8

/*
 * DVpattern defines (pattern numbers)
 */
#define MPT_DV_MAX_PATTERN     4


/****************************************************************************
 *
 * Name:    mpt_LUNComplete
 * Summary: Set the device DV status flag and print a status message
 *          Wakeup the waiting process.
 *
 * Inputs:  adapter, device, status flags and message
 * Returns: None.
 *
 * Remarks: Generate the wakeup signal when DV on this device completes.
 * 
 ****************************************************************************/
void mpt_LUNComplete (char flag, char *msg)
{
    if (DVflags & MPT_DV_DEVICE_RESERVED )
    {
        /*
         * must issue a release command....
         */
        DVcmd = MPT_DV_RELEASE;
        mpt_DVIoCmd ();
        return;
    }

    if ( flag & MPT_DV_SUCCESS)
    {
        if ( DVdp->luns == 0 )
        {
            if ( ++DVlun <= MPT_MAX_LUN )
            {
                /*
                 * issue Test Unit Ready to next lun
                 */
                DVpass = 0;
                DVflags = 0;
                DVcmd = MPT_DV_TUR;
                mpt_DVIoCmd ();
                return;
            }
        }
        else
        {
            DVflags |= MPT_DV_DONE;
            if ( DVdp->tlflags & MPT_DV_DEVICE )
            {
                ubit32 NegoData;
                NegoData = DevicePageBufPtr->RequestedParameters;
                SWAP_BYTES (NegoData);

                DVdp->NegWidth = NegoData & MPI_SCSIDEVPAGE1_RP_WIDE ? MPT_WIDE:0;
                DVdp->Offset = (NegoData & MPI_SCSIDEVPAGE1_RP_MAX_SYNC_OFFSET_MASK) >> 16;
                DVdp->SyncFactor = (NegoData & MPI_SCSIDEVPAGE1_RP_MIN_SYNC_PERIOD_MASK) >> 8;
            }
        }
     }

     if ( DVflags & MPT_DV_DEVICE_QUIESCED )
     {
         mpt_DVQuiesceEnable( MPI_RAID_ACTION_ENABLE_PHYS_IO );
         return;
     }

    if ( (flag & MPT_DV_SHOW_MSG) || (mpt_Debug_Verbose_Level > 1) )
        cmn_err (CE_WARN, "%s on ha=%d id=%d lun=%d", msg, DVpath, DVid, DVlun);

    if ( DVdp->luns == 0 )
         mpt_RemoveDevice (DVdp, 0);
    else if ( DVdp->tlflags & MPT_RAID_VOLUME )
        DVdp->Timeout =  MPT_DEFAULT_TIMEOUT;
/*    else if ( DVdp->tlflags & MPT_RAID_DEVICE )
    {
        DVdp->luns = 0;
        mpt_RemoveDevice (DVdp, 0);
    } */

    mpt_FreeCmdMFs (DVmhp, DVMFindex);
    MPT_BROADCAST(DVmhp, 0);
}

/****************************************************************************
 *
 * Name:    mpt_DeviceDiscovery
 * Summary: Perform device discovery and, if SCSI, DV on a specific device
 *
 * Inputs:  device pointer
 * Returns: None.
 *
 * Remarks: DVpass == MPT_DVPASS_CAN_RETRY indicates that this is the first run of DV. 
 * 
 ****************************************************************************/
void mpt_DeviceDiscovery (pDev_t dp)
{

    DVdp = dp;
    DVpath = DVdp->path;
    DVbus = DVdp->bus;
    DVid = DVdp->id;
    DVPhysDiskNum = DVdp->PhysDiskNum;
    DVmhp = mpt_HBAs [DVpath];
#ifdef _HPUX_SOURCE
    DVisc = DVmhp->isc;
#endif  
    DVBufferCapacity = MPT_DV_INQUIRY_LENGTH;
    DVBufferPass = 0xff;
    DVflags = 0;
    DVpass = MPT_DVPASS_CAN_RETRY;
    DVpattern = DVlun = 0;
    devNvram = DVdp->Nvram;

    if ( FirstDVBufPtr == NULL )
    {
    /*
     * Allocate Data Buffers for Inquiry commands.
     */
#ifdef _HPUX_SOURCE
        FirstDVBufPtr = (uchar *) mpt_MemAllocPhys ( MPT_DV_MAX_BUF, &FirstPhysAddr, DVisc );
	SWAP_BYTES(FirstPhysAddr);
#else 
        FirstDVBufPtr = (uchar *) mpt_MemAllocPhys ( MPT_DV_MAX_BUF, &FirstPhysAddr);
#endif

        if ( FirstDVBufPtr == NULL )
        {
             mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "Insufficient FirstDVBuf memory for DV" );
             return;
        }

#ifdef _HPUX_SOURCE
        SecondDVBufPtr = (uchar *) mpt_MemAllocPhys ( MPT_DV_MAX_BUF, &SecondPhysAddr, DVisc );
	SWAP_BYTES(SecondPhysAddr);
#else 
        SecondDVBufPtr = (uchar *) mpt_MemAllocPhys ( MPT_DV_MAX_BUF, &SecondPhysAddr);
#endif
        if ( SecondDVBufPtr == NULL )
        {
             mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "Insufficient SecondDVBuf memory for DV" );
             return;
        }
    }
    if ( DVMFtabPtr = mpt_GetFreeMF (DVmhp) )
    {
        DVMFindex=DVMFtabPtr->MFindex;
        DVMFtabPtr->dp = DVdp;
        DVMFtabPtr->Timeout =  MPT_COMMAND_TIMEOUT;
        /* 
         * The ISR will issue broadcast when completed.
         */
        DVcmd  = MPT_DV_TUR;
        if ( DVdp->tlflags & MPT_DV_DEVICE )
        {
             if ( DVdp->tlflags & MPT_RAID_DEVICE )
             {
                 mpt_DVQuiesceEnable( MPI_RAID_ACTION_QUIESCE_PHYS_IO );
             }
             else
                 mpt_DVConfigHeader ();
        }
        else
            mpt_DVIoCmd ();  /* issue TUR command */
    }
    else
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "DV Failed: No MF available" );
}

/****************************************************************************
 *
 * Name:    mpt_DVConfigHeader
 * Summary: Issue the Config Request Message to get the page size and version
 *
 * Inputs:  device pointer
 * Returns: None.
 *
 * Remarks: Construct a Config Request command for the given inputs.
 *          Add to the HBA driver_q.  Issue the command and wait for interrupt.
 *
 *          call back: DVConfigHRCompleted 
 *
 *          Hba lock held upon entry and exit.
 *
 ****************************************************************************/
void mpt_DVConfigHeader (void)
{
    Config_t    *Request;
    SGESimple32_t *mpisge;
    ubit32      vv;
    uint        ii;

    /*
     * The Reply frame has important data, so the callback will free the RF.
     * Use a generic LUN of 0 since a DVdp is defined.
     */
        MPT_DBG(MPT_DBG_DV, cmn_err(CE_CONT, "MF index 0x%x", DVMFindex));

        DVMFtabPtr->CallBackPtr = mpt_DVConfigHRCompleted;
        DVMFtabPtr->lun = 0;  
        DVMFtabPtr->Flags = MPT_TABLE_FLAGS_SAVE_REPLY_FRAME;
        DVMFtabPtr->TimeToDie = 0;

        /*
         * Populate the Message Frame 
         */
        Request= (Config_t *) DVMFtabPtr->MFVirtAddr;
        Request->Function = MPI_FUNCTION_CONFIG;
        Request->ChainOffset = 0;
        Request->Action = MPI_CONFIG_ACTION_PAGE_HEADER;
        Request->MsgFlags = 0;  
#ifdef _HPUX_SOURCE
        Request->MsgContext = DVMFindex<<8;
#else
        Request->MsgContext = DVMFindex;
#endif  

        Request->Header.PageNumber = 1;
        Request->Header.PageType =  MPI_CONFIG_PAGETYPE_SCSI_DEVICE;
        vv =  ( DVbus << 8 ) | DVid;
	SWAP_BYTES (vv);
        Request->PageAddress = vv;

        mpisge = (SGESimple32_t *) &Request->PageBufferSGE;

        vv = ( MPT_SGE_FLAGS_SIMPLE_ELEMENT | 
	       MPT_SGE_FLAGS_32_BIT_ADDRESSING |
	       MPT_SGE_FLAGS_LAST_ELEMENT |
	       MPT_SGE_FLAGS_END_OF_BUFFER |
               MPT_SGE_FLAGS_END_OF_LIST );
        SWAP_BYTES(vv);
        mpisge->FlagsLength = vv;

        mpisge->Address = 0xFFFFFFFF;
        DVMFtabPtr->CompCode = MPT_STAT_GOOD;

        mpt_StartNewCommand (DVmhp, DVMFtabPtr);
}


/****************************************************************************
 *
 * Name:    mpt_DVConfigHRCompleted 
 * Summary: Handle the callback for the Config Header Request command during 
 *          DV.
 *
 * Inputs:  Adapter structure, DVmhp, and command index
 * Returns: None.
 *
 * Remarks: The CompCode is set to the IOC Status.
 *          This routine must provide all necessary clean-up.   
 *
 ****************************************************************************/
void mpt_DVConfigHRCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    SCSIDevicePage1_t *bufPtr;
    ubit32 CompCode, physAddr, length;
    ConfigReply_t *Reply;

    CompCode = MFtabPtr->CompCode;
    physAddr = MFtabPtr->ReplyPhysAddr;

    if (CompCode == MPI_IOCSTATUS_SUCCESS)
    {
        /*
         * Save the reply header, allocate memory for the page buffer
         * and issue the read page. Otherwise, free memory and exit DV. 
         */
        Reply = (ConfigReply_t *) ( (ulong_t)(physAddr - DVmhp->ReplyBasePhysAddr) + DVmhp->ReplyBaseVirtAddr);
        DVDevicePageHeader = Reply->Header;
        DVmhp->PageLength = Reply->Header.PageLength;  /* FIX */
        DVmhp->PageVersion = Reply->Header.PageVersion;

        mpt_PostRFtoIOC (DVmhp, physAddr);

	length = DVDevicePageHeader.PageLength * 4;
#ifdef _HPUX_SOURCE
        bufPtr = (SCSIDevicePage1_t *) mpt_MemAllocPhys ( (ubit32)length, &DevicePagePhysAddr, DVisc);
#else
        bufPtr = (SCSIDevicePage1_t *) mpt_MemAllocPhys ( (ubit32)length, &DevicePagePhysAddr);
#endif
	if ( bufPtr == NULL )
        {
            mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DV Failed: DevicePageBuf MemAllocPhys Failed" );
            return;
        }

	DevicePageBufPtr = bufPtr;
        /*
         * Primary effect of this read is to prepare for the 
         * upcoming writes - as all config page chages are made
         * with a read modify write algorithm.
        */
        mpt_DVReadPage ();
    }
    else if ((CompCode == MPT_DID_TIMEOUT) || (CompCode == MPT_DID_IOC_HARD_RESET))
    {
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "DV Failed: Device Timeout or Hard Reset" );
    }
    else
    {
        mpt_PostRFtoIOC (DVmhp, physAddr);
        MPT_DBG(MPT_DBG_DV, cmn_err(CE_CONT, "IocStatus 0x%x", CompCode ));
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DV Failed: Config Error" );
    }
}

/****************************************************************************
 *
 * Name:    mpt_DVReadPage
 * Summary: Issue the Message to read the page header 
 *
 * Inputs:  device pointer
 * Returns: None.
 *
 * Remarks: Construct an Config command for the given inputs.
 *          Add to the HBA driver_q.  Issue the command and wait for interrupt.
 *
 *          call back: DVReadPageCompleted 
 *
 *          Hba lock held upon entry and exit.
 *
 ****************************************************************************/
void mpt_DVReadPage (void)
{
    Config_t    *Request;
    SGESimple32_t *mpisge;
    ubit32      vv;
    uint        ii;

        MPT_DBG(MPT_DBG_DV, cmn_err(CE_CONT, "MF index 0x%x", DVMFindex));
        /*
         * Populate the TableStruct_t Use a generic LUN of 0
         * since DVdp is defined.
         */
        DVMFtabPtr->CallBackPtr = mpt_DVReadPageCompleted;
        DVMFtabPtr->lun = 0;
        DVMFtabPtr->Flags = 0;
        DVMFtabPtr->TimeToDie = 0;

        /*
         * Populate the Message Frame 
         */
        Request= (Config_t *) DVMFtabPtr->MFVirtAddr;
        Request->Function = MPI_FUNCTION_CONFIG;
        Request->ChainOffset = 0;
        Request->Action = MPI_CONFIG_ACTION_PAGE_READ_CURRENT;
        Request->MsgFlags = 0;  
#ifdef _HPUX_SOURCE
        Request->MsgContext = DVMFindex<<8;
#else
        Request->MsgContext = DVMFindex;
#endif  

        Request->Header = DVDevicePageHeader;
        vv =  ( DVbus << 8 ) | DVid;
	SWAP_BYTES (vv);
        Request->PageAddress = vv;

        mpisge = (SGESimple32_t *) &Request->PageBufferSGE;

        vv = ( MPT_SGE_FLAGS_SIMPLE_ELEMENT | 
	       MPT_SGE_FLAGS_32_BIT_ADDRESSING |
	       MPT_SGE_FLAGS_LAST_ELEMENT |
	       MPT_SGE_FLAGS_END_OF_BUFFER |
               MPT_SGE_FLAGS_END_OF_LIST ) |
               DVDevicePageHeader.PageLength * 4;
        SWAP_BYTES(vv);
        mpisge->FlagsLength = vv;

        vv = DevicePagePhysAddr;
        SWAP_BYTES(vv);
        mpisge->Address = vv;
        DVMFtabPtr->CompCode = MPT_STAT_GOOD;

        mpt_StartNewCommand (DVmhp, DVMFtabPtr);
}


/****************************************************************************
 *
 * Name:    mpt_DVReadPageCompleted 
 * Summary: Handle the callback for the Read Page Request command during 
 *          DV.
 *
 * Inputs:  Adapter structure, DVmhp, and command index
 * Returns: None.
 *
 * Remarks: The CompCode is set to the IOC Status.
 *          This routine must provide all necessary clean-up.   
 *
 ****************************************************************************/
void mpt_DVReadPageCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    ubit32 CompCode;

    CompCode = MFtabPtr->CompCode;

    if (CompCode == MPI_IOCSTATUS_SUCCESS)
    {
        mpt_DVParms ( MPT_DV_GET_NVRAM_PARMS);
        DVpass = 0;
        mpt_DVParms ( MPT_DV_SET_MIN);
    }
    else if ((CompCode == MPT_DID_IOC_HARD_RESET) || ( CompCode == MPT_DID_TIMEOUT || CompCode == MPT_STAT_SBUSER ))
    {
        DVpass = 0;
        mpt_DVParms ( MPT_DV_SET_FALLBACK);
    }
    else
    {
        MPT_DBG(MPT_DBG_DV, cmn_err(CE_CONT, "IocStatus 0x%x", CompCode ));
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DVReadPage Failed: Config Error" );
    }
}

/****************************************************************************
 *
 * Name:    mpt_DVWritePage
 * Summary: Issue the Message to write the page 
 *
 * Inputs:  device pointer
 * Returns: None.
 *
 * Remarks: Construct a Config command for the given inputs.
 *          Add to the HBA driver_q.  Issue the command and wait for interrupt.
 *
 *          call back: DVWritePageCompleted 
 *
 *          Hba lock held upon entry and exit.
 *
 *          If DVpass = -1, do a write commit.
 *
 ****************************************************************************/
void mpt_DVWritePage (void)
{
    Config_t    *Request;
    SGESimple32_t *mpisge;
    ubit32      vv;

        MPT_DBG(MPT_DBG_DV, cmn_err(CE_CONT, "MF index 0x%x", DVMFindex));

        /*
         * Populate the TableStruct_t
         * Use a generic LUN of 0 since DVdp is defined.
         */
        DVMFtabPtr->CallBackPtr = mpt_DVWritePageCompleted;
        DVMFtabPtr->lun = 0;
        DVMFtabPtr->Flags = 0;
        DVMFtabPtr->TimeToDie = 0;

        /*
         * Populate the Message Frame 
         */
        Request = (Config_t *) DVMFtabPtr->MFVirtAddr;
        Request->Function = MPI_FUNCTION_CONFIG;
        Request->ChainOffset = 0;
        Request->Action = MPI_CONFIG_ACTION_PAGE_WRITE_CURRENT;
        Request->MsgFlags = 0;  

#ifdef _HPUX_SOURCE
        Request->MsgContext = DVMFindex<<8;
#else
        Request->MsgContext = DVMFindex;
#endif  

        Request->Header = DVDevicePageHeader;
        vv =  ( DVbus << 8 ) | DVid;
	SWAP_BYTES (vv);
        Request->PageAddress = vv;

        mpisge = (SGESimple32_t *) &Request->PageBufferSGE;

        vv = ( MPT_SGE_FLAGS_SIMPLE_ELEMENT | 
	       MPT_SGE_FLAGS_32_BIT_ADDRESSING |
               MPT_SGE_FLAGS_HOST_TO_IOC |
	       MPT_SGE_FLAGS_LAST_ELEMENT |
	       MPT_SGE_FLAGS_END_OF_BUFFER |
               MPT_SGE_FLAGS_END_OF_LIST ) |
               DVDevicePageHeader.PageLength * 4;
        SWAP_BYTES(vv);
        mpisge->FlagsLength = vv;

        vv = DevicePagePhysAddr;
        SWAP_BYTES(vv);
        mpisge->Address = vv;
        DVMFtabPtr->CompCode = MPT_STAT_GOOD;

        mpt_StartNewCommand (DVmhp, DVMFtabPtr);
}


/****************************************************************************
 *
 * Name:    mpt_DVWritePageCompleted 
 * Summary: Handle the callback for the Write Page Request command during 
 *          DV.
 *
 * Inputs:  Adapter structure, DVmhp, and command index
 * Returns: None.
 *
 * Remarks: The CompCode is set to the IOC Status.
 *          This routine must provide all necessary clean-up.   
 *
 ****************************************************************************/
void mpt_DVWritePageCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    ubit32 CompCode;
    ubit32 NegoData;
    uchar Width, Offset;

    CompCode = MFtabPtr->CompCode;

    if (CompCode == MPI_IOCSTATUS_SUCCESS)
    {
        /*
         * Good! 
         * If DV enabled, issue the next command.
         * Else, call the complete function
         */
        if (mpt_DoDomainValidation == 0) 
        {
#ifdef MPT_DEBUG
            flags = MPT_DV_SHOW_MSG | MPT_DV_SUCCESS;
#else
            flags = MPT_DV_SUCCESS;
#endif
            mpt_LUNComplete ( flags, "DV Disabled" );
	    return;
        }

        if (DVpass & MPT_DVPASS_MASK)
	{
            NegoData = DevicePageBufPtr->RequestedParameters;
	    SWAP_BYTES (NegoData);
            Width = NegoData & MPI_SCSIDEVPAGE1_RP_WIDE ? MPT_WIDE:0;
            Offset = (NegoData & MPI_SCSIDEVPAGE1_RP_MAX_SYNC_OFFSET_MASK) >> 16; 

            if ( Offset == 0 && Width == 0 && (DVflags & MPT_DV_INQ_DONE) )
            {
                mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_SUCCESS, "DV Success: Async/Narrow");
                return;
            }
	}
        mpt_DVIoCmd ();
    }
    else if ((CompCode == MPT_DID_IOC_HARD_RESET) || ( CompCode == MPT_DID_TIMEOUT || CompCode == MPT_STAT_SBUSER ))
    {
        DVpass = 0;
        mpt_DVParms ( MPT_DV_SET_FALLBACK);
    }
    else
    {
        /*
         * Write Failed. Exit. 
         */
        MPT_DBG(MPT_DBG_DV, cmn_err(CE_CONT, "IocStatus 0x%x", CompCode ));
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DVWritePage Failed: Config Error" );
    }
}

/****************************************************************************
 *
 * Name:    mpt_DVQuiesceEnable
 * Summary: Issue the Request Message to quiesce or enable a Raid Phys Disk
 *
 * Inputs:  None
 * Returns: None.
 *
 * Remarks: Construct a Quiese or Enable Request for the given inputs.
 *          Add to the HBA driver_q.  Issue the command and wait for interrupt.
 *
 *          call back: DVQuiesceCompleted or DVEnableCompleted
 *
 *          Hba lock held upon entry and exit.
 *
 ****************************************************************************/
void mpt_DVQuiesceEnable (uchar Action )
{
    MpiRaidActionRequest_t *Request;
    SGESimple32_t *mpisge;
    ubit32      vv;
    uint        ii;
    uchar       DVInProgress;

    DVMFtabPtr->lun = 0;  
    DVMFtabPtr->Flags = 0;
    DVMFtabPtr->TimeToDie = 0;

    /*
     * Populate the Message Frame 
     */
    Request= (MpiRaidActionRequest_t *) DVMFtabPtr->MFVirtAddr;
    Request->Function = MPI_FUNCTION_RAID_ACTION;
    Request->ChainOffset = 0;
    Request->Action = Action;
    Request->MsgFlags = 0;  
#ifdef _HPUX_SOURCE
    Request->MsgContext = DVMFindex<<8;
#else
    Request->MsgContext = DVMFindex;
#endif  

    Request->PhysDiskNum = DVPhysDiskNum;
    Request->VolumeBus = DVmhp->IOCPage2Ptr->RaidVolume[0].VolumeBus;
    Request->VolumeID = DVmhp->IOCPage2Ptr->RaidVolume[0].VolumeID;

    DVMFtabPtr->CompCode = MPT_STAT_GOOD;

    if ( Action == MPI_RAID_ACTION_QUIESCE_PHYS_IO )
    {
        DVInProgress = ( DVmhp->HBA_Chars & MPT_DV_PHYSDISK_IN_PROGRESS ); 
        DVmhp->HBA_Chars |= MPT_DV_PHYSDISK_IN_PROGRESS;  /* PhysDisk DV in progress */
        DVMFtabPtr->CallBackPtr = mpt_DVQuiesceCompleted;
        DVMFtabPtr->Flags |= MPT_TABLE_FLAGS_QUIESCE_COMMAND;

        if ( DVInProgress )  /* Is a PhysDisk DV already in progress */
        {
            Q_APPEND ( ((QHEADPtr)(&DVmhp->ready_q)), ((QHEADPtr)(DVMFtabPtr)) ); 
        }
        else
            mpt_StartNewCommand (DVmhp, DVMFtabPtr);
    }
    else
    {
        DVMFtabPtr->CallBackPtr = mpt_DVEnableCompleted;
        mpt_StartNewCommand (DVmhp, DVMFtabPtr);
    }
}

/****************************************************************************
 *
 * Name:    mpt_DVParms
 * Summary: Write/Save negotiation data to the saved device page
 *          On "SET" operations, call the WritePage routine.
 *      
 *
 * Inputs:  device pointer and flag that contains control information.
 * Returns: None.
 *
 * Remarks: If DVpass == -1, do a write commit
 *
 * Factor Width Period   MT/sec MB/sec Label
 *  0x08    1     6.25    160    320   ULTRA4 & WIDE
 *  0x09    1    12.5      80    160   ULTRA3 & WIDE
 *  0x0A    1    25.0      40     80   ULTRA2 & WIDE
 *  0x0A    0    25.0      40     40   ULTRA2 & NARROW
 *  0x0C    1    48.0      20     40   ULTRA  & WIDE
 *  0x0C    0    48.0      20     20   ULTRA  & NARROW
 *  0x19    1   100.0      10     20   FAST   & WIDE
 *  0x19    0   100.0      10     10   FAST   & NARROW
 *  0x32    1   200.0       5     10   SCSI   & WIDE
 *  0x32    0   200.0       5      5   SCSI   & NARROW
 * 
 *  If offset == 0,         5     10   ASYNC  & WIDE
 *  If offset == 0,         5      5   ASYNC  & NARROW
 ****************************************************************************/
void mpt_DVParms ( uchar flag)
{
    ubit32 NegoData, vv;
    uchar Width, Offset, SyncFactor, NVSyncFactor;

    if (flag == MPT_DV_SET_MIN)
    {
	Width = Offset = 0;
        SyncFactor = MPT_FAST;
    }
    else 
    {
      NegoData = DevicePageBufPtr->RequestedParameters;
      SWAP_BYTES(NegoData);

      if (flag == MPT_DV_SET_FALLBACK)
      {
          if (DVBufferPass == MPT_DVPASS_ECHO_BUFFER )
          {
                DVBufferPass = MPT_DVPASS_DATA_BUFFER;
                DVcmd = MPT_DV_READ_BUFFER_CAPACITY;
                mpt_DVIoCmd ();
                return;
          }
        /* 
         * Get last values that were requested 
         */
        Width = NegoData & MPI_SCSIDEVPAGE1_RP_WIDE ? MPT_WIDE:0;
        Offset = (NegoData & MPI_SCSIDEVPAGE1_RP_MAX_SYNC_OFFSET_MASK) >> 16;
        SyncFactor = (NegoData & MPI_SCSIDEVPAGE1_RP_MIN_SYNC_PERIOD_MASK) >> 8;

        if (Offset)
        {
            if (SyncFactor < MPT_ULTRA3)
            {
                /*
                 *  U320 to U160
                 */
                SyncFactor = MPT_ULTRA3;
            }
            else if (SyncFactor < MPT_ULTRA2)
            {
                /*
                 *  U160 to Ultra2 Wide
                 */
                SyncFactor = MPT_ULTRA2;
            }
            else if (SyncFactor < MPT_ULTRA)
            {
                if (Width)
                {
                    /*
                     *  Ultra2 Wide to Ultra2 Narrow
                     */
                    Width = 0;
                }
                else
                {
                    /*
                     *  Ultra2 Narrow to Ultra Wide (if allowed)
                     */
                    SyncFactor = MPT_ULTRA;
                    Width = DVdp->Width;
                }
            }
            else if (SyncFactor < MPT_FAST)
            {
                if (Width)
                {
                    /*
                     *  Ultra Wide to Ultra Narrow
                     */
                    Width = 0;
                }
                else
                {
                    /*
                     *  Ultra Narrow to Fast Wide (if allowed)
                     */
                    SyncFactor = MPT_FAST;
                    Width = DVdp->Width;
                }
            }
            else if (SyncFactor < MPT_SCSI)
            {
                if (Width)
                {
                    /*
                     *  Fast Wide to Fast Narrow 
                     */
                    Width = 0;
                }
                else
                {
                    /*
                     *  Fast Narrow to 5MT/sec Wide (if allowed)
                     */
                    SyncFactor = MPT_SCSI;
                    Width = DVdp->Width;
                }
            }
            else 
            {
                if (Width)
                {
                    /*
                     *  5MT/sec Narrow to 5MT/sec Narrow 
                     */
                    Width = 0;
                }
                else
                {
                    /*
                     *  5MT/sec sync Narrow to Async Wide
                     */
                    Offset = 0;
                    Width = DVdp->Width;
                }
            }
        }
        else if (Width)
        {
            /*
             * Async Wide to Async Narrow
             */
            Width = 0;
        }
        DVpattern = 0;
      }
      else if (flag == MPT_DV_GET_NVRAM_PARMS)
      {
        Width = 1;
        Offset = DVmhp->Offset;
        SyncFactor = DVmhp->SyncFactor;
        /*
         * Verify that NVRAM data has been obtained for this target 
         * (always bus 0). If no NVRAM, use adapter settings.
         */
        if ( devNvram == MPT_NVRAM_UNKNOWN)
        {
            DVdp->Timeout = MPT_DEFAULT_TIMEOUT;
        }
        else
        {
            if ( (DVdp->Timeout = (uchar)devNvram) == 0 )
                DVdp->Timeout = 0xffffffff;
            /*
             * Width
             */
            if ( devNvram & MPT_NVRAM_WIDE_DISABLE )
            {
                Width = 0;
                DVdp->tlflags |= MPT_NO_WDTR;
            }
    
            if ( (devNvram & MPT_NVRAM_DISCONNECT_ENABLE) == 0 )
                DVdp->Control |= MPI_SCSIIO_CONTROL_NO_DISCONNECT;

            /*
             * Set offset only if this device can handle sync. 
             * (NVRAM reports a 0 sync factor for async. devices).
             */
            if (devNvram & MPT_NVRAM_SYNC_MASK)
            {
                NVSyncFactor = (devNvram & MPT_NVRAM_SYNC_MASK)>> MPT_NVRAM_SYNC_SHIFT;
                if ( NVSyncFactor > SyncFactor )
                    SyncFactor = NVSyncFactor;
            }
            else
            {
                Offset = 0;
                SyncFactor = MPT_FAST;
                DVdp->tlflags |= MPT_NO_SDTR;
            }
            if ( devNvram & MPT_NVRAM_TAG_QUEUE_ENABLE )
        	DVdp->Control |= MPI_SCSIIO_CONTROL_SIMPLEQ;
            else
            	DVdp->Control |= MPI_SCSIIO_CONTROL_UNTAGGED;
        }

        if ( (DVdp->NegWidth = Width) == 0 )
        {
            if ( SyncFactor < MPT_ULTRA2 )
                 SyncFactor = MPT_ULTRA2;
        }
        DVdp->SyncFactor = SyncFactor;
        DVdp->Offset = Offset;
        MPT_DBG(MPT_DBG_DV, cmn_err(CE_CONT, "Device: Width 0x%x, SyncFactor 0x%x, Offet 0%x\n", Width, SyncFactor, Offset));
        MPT_DBG(MPT_DBG_DV, cmn_err(CE_CONT, "Adapter: SyncFactor 0x%x, Offet 0%x\n", DVmhp->SyncFactor, DVmhp->Offset));

        return;
      }
      else if (flag == MPT_DV_SET_MAX)
      {
        SyncFactor = (DVmhp->SyncFactor < DVdp->SyncFactor)? DVdp->SyncFactor : DVmhp->SyncFactor;

        if ( (Width = DVdp->Width) == 0 )
        {
            if ( SyncFactor < MPT_ULTRA2 )
                 SyncFactor = MPT_ULTRA2;
        }

        Offset = DVdp->Offset;
      }
      DVdp->NegWidth = Width;
      DVdp->SyncFactor = SyncFactor;
      DVdp->Offset = Offset;
    }
    NegoData = (Offset << 16) | (SyncFactor << 8 );
    vv = 0;
    if ( Width )
    {
        NegoData |= MPI_SCSIDEVPAGE1_RP_WIDE;
        if (Offset && SyncFactor <= MPT_ULTRA3)
        {
            if ( SyncFactor == MPT_ULTRA4 )
                NegoData |= ( MPI_SCSIDEVPAGE1_RP_IU | 
                              MPI_SCSIDEVPAGE1_RP_DT |
                              MPI_SCSIDEVPAGE1_RP_QAS );
            else
                NegoData |= MPI_SCSIDEVPAGE1_RP_DT;
        }
    }
    else
    {
        if ( DVdp->tlflags & MPT_NO_WDTR )
            vv = MPI_SCSIDEVPAGE1_CONF_WDTR_DISALLOWED;
    }
    if ( DVdp->tlflags & MPT_NO_SDTR )
        vv |= MPI_SCSIDEVPAGE1_CONF_SDTR_DISALLOWED;
    SWAP_BYTES(vv);
    DevicePageBufPtr->Configuration = vv;

    SWAP_BYTES(NegoData);
    DevicePageBufPtr->RequestedParameters = NegoData;
    mpt_DVWritePage ();
}

/****************************************************************************
 * Name:    mpt_DVIoCmd
 * Summary: Issue a DV IO command 
 *          TUR, INQ, RESERVE, READ BUFF, WRITE BUFF, RELEASE
 *
 * Inputs:  device pointer, command code
 * Returns: None.
 *
 * Remarks: Construct the specified IO command.
 *          Add to the HBA driver_q.  
 *          Issue the command and wait for callback:
 *          a) Interrupt -> command completed; selection timeout 
 *                          device busy; other
 *          b) Timeout  ->  Quit the bus scan, remove all devices, and
 *                          try again if retry_count is small enough.
 *
 *          Hba lock held upon entry and exit.
 ****************************************************************************/
void mpt_DVIoCmd (void)
{
    SCSIIORequest_t *Request;
    SGESimple32_t *mpisge;
    ubit32      bufferPhysAddr=0xFFFFFFFF;
    ubit32      dataLength=0;
    ubit32      vv;
    uint        ii;
    uchar cmd[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    uchar cmdSize = MPT_6BYTE_CMD;

        cmd[0] = DVcmd;
        for (ii = 1; ii < 16; ii++)
            cmd[ii] = 0;

        DVMFtabPtr->Flags = 0;
        DVMFtabPtr->TimeToDie = 0;

        MPT_DBG(MPT_DBG_DV, cmn_err(CE_CONT, "MF index 0x%x", DVMFindex));
        Request = (SCSIIORequest_t *) DVMFtabPtr->MFVirtAddr;
        Request->Control = MPI_SCSIIO_CONTROL_NO_DISCONNECT | MPI_SCSIIO_CONTROL_UNTAGGED;

        /*
         * For each command:
         *  1) set the callback pointer
         *  2) update cmd
         *  3) set the physical address for the IO buffer
         */

        if (DVcmd == MPT_DV_INQUIRY)
        {
            Request->Control |= MPI_SCSIIO_CONTROL_READ;
            if ( DVflags & MPT_DV_DONE )
            {
                DVMFtabPtr->CallBackPtr = mpt_DVInqLunCompleted;
                bufferPhysAddr = FirstPhysAddr;
            }
            else
            {
                DVMFtabPtr->CallBackPtr = mpt_DVInqCompleted;
                if ( (DVpass & MPT_DVPASS_MASK) < 2)
                    bufferPhysAddr = FirstPhysAddr;
                else 
                    bufferPhysAddr = SecondPhysAddr;
            }

            cmd[4] = MPT_DV_INQUIRY_LENGTH;

            DVInqLength = dataLength = MPT_DV_INQUIRY_LENGTH;
        }
        else if (DVcmd == MPT_DV_WRITE_BUFFER)
        {
            Request->Control |= MPI_SCSIIO_CONTROL_WRITE;
            DVMFtabPtr->CallBackPtr = mpt_DVWriteBufferCompleted;

            cmdSize = MPT_10BYTE_CMD;
            /*
             * Select echo data buffer (0x0a) or data buffer (0x02)
             *  echo descriptor = 1010 b = 0x0A
             *       descriptor = 0010 b = 0x02
             */

            dataLength = DVBufferCapacity;
            cmd[1] = 0x02 | DVBufferPass;
            cmd[6] = (dataLength >> 16) & 0xFF;
            cmd[7] = (dataLength >> 8) & 0xFF;
            cmd[8] = dataLength & 0xFF;

            bufferPhysAddr = FirstPhysAddr;
        }
        else if (DVcmd == MPT_DV_READ_BUFFER)
        {
            Request->Control |= MPI_SCSIIO_CONTROL_READ;
            DVMFtabPtr->CallBackPtr = mpt_DVReadBufferCompleted;

            cmdSize = MPT_10BYTE_CMD;
            /*
             * Select echo data buffer (0x0a) or data buffer (0x02)
             *  echo descriptor = 1010 b = 0x0A
             *       descriptor = 0010 b = 0x02
             */

            dataLength = DVBufferCapacity;
            cmd[1] = 0x02 | DVBufferPass;
            cmd[6] = (dataLength >> 16) & 0xFF;
            cmd[7] = (dataLength >> 8) & 0xFF;
            cmd[8] = dataLength & 0xFF;

            bufferPhysAddr = SecondPhysAddr;
        }
        else if (DVcmd == MPT_DV_READ_BUFFER_CAPACITY)
        {
            Request->Control |= MPI_SCSIIO_CONTROL_READ;
            /*
             * Specialized - read capacity of the (echo) buffer first.
             */
            DVMFtabPtr->CallBackPtr = mpt_DVReadBufferCapCompleted;

            cmdSize = MPT_10BYTE_CMD;
            cmd[0] = MPT_DV_READ_BUFFER;
            /*
             * Select echo data buffer (0x0B) or data buffer (0x03)
             *  echo descriptor = 1011 b = 0x0B
             *       descriptor = 0011 b = 0x03
             */

            cmd[1] = 0x03 | DVBufferPass;
            cmd[8] = MPT_DV_READ_BUFFER_CAP_LENGTH;
            dataLength = MPT_DV_READ_BUFFER_CAP_LENGTH;

            bufferPhysAddr = FirstPhysAddr;
        }
        else if (DVcmd == MPT_DV_TUR)
        {
            DVMFtabPtr->CallBackPtr = mpt_DVTURCompleted;
        }
        else if (DVcmd == MPT_DV_RESERVE)
        {
            DVMFtabPtr->CallBackPtr = mpt_DVReserveCompleted;
        }
        else if ((DVcmd == MPT_DV_RELEASE) || (DVcmd == MPT_DV_RELEASE_THEN_RESERVE) )
        {
            DVMFtabPtr->CallBackPtr = mpt_DVReleaseCompleted;

            cmd[0] = MPT_DV_RELEASE;
        }
        else if (DVcmd == MPT_DV_START)
        {
            DVMFtabPtr->CallBackPtr = mpt_DVTURCompleted;

            cmd[4] = 0x01;
        }
        else
        {
            /*
             * Unknown or unsupported DV command. Just exit.
             */
	    cmn_err (CE_WARN," Unknown DVIoCmd=%x", DVcmd);
            return;
        }
        /*
         * Populate the TableStruct_t
         */
        DVMFtabPtr->lun = DVlun;

        /*
         * Populate the Message Frame 
         */
        Request->Bus = DVbus;
        Request->ChainOffset = 0;
        if ( DVdp->tlflags & MPT_RAID_DEVICE )
        {
            Request->Function = MPI_FUNCTION_RAID_SCSI_IO_PASSTHROUGH;
            Request->TargetID = DVPhysDiskNum;
        }
        else
        {
            Request->Function = MPI_FUNCTION_SCSI_IO_REQUEST;
            Request->TargetID = DVid;
        }
        Request->CDBLength = cmdSize;
        Request->SenseBufferLength = MPT_SENSE_BUFFER_SIZE;
        Request->MsgFlags = 0;    /* 32 bit sense buffer*/  
#ifdef _HPUX_SOURCE
        Request->MsgContext = DVMFindex<<8;
#else
        Request->MsgContext = DVMFindex;
#endif  

        for (ii=0; ii<8; ii++)
            Request->LUN[ii] = 0;

        Request->LUN[1] = DVlun;

        for (ii=0; ii < 16; ii++)
            Request->CDB[ii] = cmd[ii];

        vv = dataLength;
        SWAP_BYTES(vv);
        Request->DataLength = vv;
        Request->SenseBufferLowAddr = DVMFtabPtr->SensePhysAddr;

        mpisge = (SGESimple32_t *) &Request->SGL;

        if ( Request->Control & MPI_SCSIIO_CONTROL_WRITE)
            vv = ( MPT_SGE_FLAGS_SIMPLE_ELEMENT | 
                   MPT_SGE_FLAGS_32_BIT_ADDRESSING |
                   MPT_SGE_FLAGS_HOST_TO_IOC |
                   MPT_SGE_FLAGS_LAST_ELEMENT |
                   MPT_SGE_FLAGS_END_OF_BUFFER |
                   MPT_SGE_FLAGS_END_OF_LIST ) |
                   dataLength;
        else
            vv = ( MPT_SGE_FLAGS_SIMPLE_ELEMENT | 
                   MPT_SGE_FLAGS_32_BIT_ADDRESSING |
                   MPT_SGE_FLAGS_LAST_ELEMENT |
                   MPT_SGE_FLAGS_END_OF_BUFFER |
                   MPT_SGE_FLAGS_END_OF_LIST ) |
                   dataLength;
        SWAP_BYTES(vv);
        mpisge->FlagsLength = vv;

        mpisge->Address = bufferPhysAddr;
        DVMFtabPtr->CompCode = MPT_STAT_GOOD;

        mpt_StartNewCommand (DVmhp, DVMFtabPtr);
}


/****************************************************************************
 *
 * Name:    mpt_DVTURCompleted 
 * Summary: Handle the callback for the Test Unit Ready command during DV.
 *
 * Inputs:  Adapter structure, DVmhp, and command index
 * Returns: None.
 *
 ****************************************************************************/
void mpt_DVTURCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    ubit32 CompCode;

    CompCode = MFtabPtr->CompCode;

    if (CompCode == MPT_STAT_NOTEQ)
    {
#ifdef MPT_DEBUG
        flags = MPT_DV_SHOW_MSG | MPT_DV_FAIL;
#else
        flags = MPT_DV_FAIL;
#endif
        mpt_LUNComplete ( flags, "DVTUR Failed: Selection Timeout" );
    }
    else if (CompCode == MPT_STAT_GOOD)
    {
        DVcmd = MPT_DV_INQUIRY;
        mpt_DVIoCmd ();
    }
    else if (CompCode == MPT_STAT_CKSTAT)
    {
        /* 
         * Check condition.
         */
        if (DVdp->SenseData[2] == MPT_SK_UNATTEN)
        {
            mpt_DVIoCmd ();  /* Retry TUR */
            return;
        }
	else if ( (DVdp->SenseData[2] == MPT_SK_ILLREQ) && (DVdp->SenseData[12] ==  MPT_ASC_INVLUN) )
	{
#ifdef MPT_DEBUG
            flags = MPT_DV_SHOW_MSG | MPT_DV_SUCCESS;
#else
            flags = MPT_DV_SUCCESS;
#endif
            mpt_LUNComplete ( flags, "DVTUR Success: Invalid LUN" );
            return;
        }
	else if ( (DVdp->SenseData[2] == MPT_SK_NREADY) && (DVdp->SenseData[12] ==  MPT_ASC_DRVNTRDY) )
        {
	    if ( DVdp->SenseData[13] == MPT_ASCQ_START_NEEDED )
	    {
                DVcmd = MPT_DV_START;
                mpt_DVIoCmd ();
                return;
            }
	    else if ( DVdp->SenseData[13] == MPT_ASCQ_START_IN_PROGRESS )
	    {
                MPT_TIMER(1000000);  /* Wait a second */
                mpt_DVIoCmd ();  /* Retry TUR */
                return;
            }
        }

        DVcmd = MPT_DV_INQUIRY;
        if ( DVdp->tlflags & MPT_DV_DEVICE )
        {
            if ( DVdp->tlflags & MPT_RAID_DEVICE )
            {
                mpt_DVQuiesceEnable( MPI_RAID_ACTION_QUIESCE_PHYS_IO );
            }
            else
                mpt_DVConfigHeader ();
        }
        else
            mpt_DVIoCmd ();
        return;
    }
    else if (CompCode == MPT_DID_TIMEOUT)
    {
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "DVTUR Failed: Timeout" );
    }
    else if (CompCode == MPT_DID_IOC_HARD_RESET)
    {
        /*
         * Fatal Error.
         */
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "DVTUR Failed: Hard Reset" );
    }
    else
    {
        if (CompCode == MPT_STAT_RETRY)
        {
            cmn_err (CE_NOTE, "DVTUR: Retry on ha=%d id=%d lun=%d\n", DVpath, DVid, DVlun);
            mpt_DVIoCmd ();
        }
        else if (CompCode == MPT_STAT_HAERR)
        {
            if ( MFtabPtr->Flags & MPT_TABLE_FLAGS_RSP_INV_CMD_FIELDS )
	    {
#ifdef MPT_DEBUG
                flags = MPT_DV_SHOW_MSG | MPT_DV_SUCCESS;
#else
                flags = MPT_DV_SUCCESS;
#endif
                mpt_LUNComplete ( flags, "DVTUR Success: Invalid LUN" );
            }
            else
                mpt_LUNComplete ( MPT_DV_FAIL, "DVTUR Failed" );
        }
        else
            mpt_LUNComplete ( MPT_DV_FAIL, "DVTUR Failed" );
    }
}


/****************************************************************************
 *
 * Name:    mpt_DVInqCompleted 
 * Summary: Handle the callback for the Inquiry command during the 
 *          bus scan.
 *
 * Inputs:  Adapter structure, DVmhp, and command pointer, MFtabPtr
 * Returns: None.
 *
 * Remarks: The scsi status and the completion code have been set by the ISR.  
 *          The Inquiry data has not been copied to the Dev_t structure.
 *          This routine must provide all necessary clean-up.   
 *
 ***************************************************************************/
void mpt_DVInqCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    SCSIIORequest_t *ioreq;
    ubit32 CompCode, vv;

    CompCode = MFtabPtr->CompCode;

    if ( (CompCode == MPT_STAT_GOOD) || ( CompCode == MPT_STAT_UNDERRUN ) )
    {
        /*
         * If pass 0 or 1, just repeat the inquiry command.
         * If pass 2, negotiate for Maximum 
         * If pass > 2, compare the buffers
         */
        if ( (DVpass & MPT_DVPASS_MASK) == 0)
        {
            DVpass++;
            mpt_DVIoCmd ();
        }
        else if ((DVpass & MPT_DVPASS_MASK) == 1)
        {
            mpt_setDevParameters ( FirstDVBufPtr);
            DVpass++;
            if ( DVdp->tlflags & MPT_DV_DEVICE )
            {
                mpt_DVParms ( MPT_DV_SET_MAX);
            }
            else
            {
                mpt_DVIoCmd ();  /* if FIBRE or RAID VOLUME, issue Inquiry again */
            }
        }
        else 
	{
    	    ioreq = (SCSIIORequest_t *)MFtabPtr->MFVirtAddr;
            vv = ioreq->DataLength;
            SWAP_BYTES ( vv );
	    if ( vv < DVInqLength )
	        DVInqLength = vv;
	    if ( DVInqLength && bcmp ((caddr_t)FirstDVBufPtr, (caddr_t)SecondDVBufPtr, DVInqLength))
            {
                 if ( DVdp->tlflags & MPT_DV_DEVICE )
                 {
                    /*
                     * Miscompare 
                     *    if this is the first go around start all over again. 
                     *          (nego async/narrow first).
                     *    else do a fallback and start all over again.
                     */
                    if ( (DVpass & MPT_DVPASS_CAN_RETRY))
                    {
                        DVpass = 0;
                        mpt_DVParms ( MPT_DV_SET_MIN);
                    }
                    else 
                    {
                        DVpass = 3;
                        mpt_DVParms ( MPT_DV_SET_FALLBACK);
                    }
                }
                else
                {
                     mpt_setDevParameters ( SecondDVBufPtr);
                     if ( DVdp->tlflags & MPT_RAID_VOLUME )
                         mpt_LUNComplete ( MPT_DV_SUCCESS, "RAID VOLUME Inq Success" );
                     else
                         mpt_LUNComplete ( MPT_DV_SUCCESS, "FIBRE Inq Success" );
                }
            }
            else
            {
                /*
                 * We had a compare,
                 * proceed to the read/write buffer tests.
                 */
/*                if (FirstDVBufPtr)
                     kmem_free (FirstDVBufPtr, DVBufferCapacity);
                  if (SecondDVBufPtr)
                     kmem_free (SecondDVBufPtr, DVBufferCapacity); */
    
                 mpt_setDevParameters ( FirstDVBufPtr);

                 if ( DVdp->tlflags & MPT_DV_DEVICE )
                 {
                     if (DVdp->SyncFactor == MPT_ULTRA4)
                         mpt_ReadConfigPage ( DVmhp, 0, MPI_CONFIG_PAGETYPE_SCSI_DEVICE, (DVbus<<8 | DVid) );  /* Fujitsu, Quantum workaround */
                     else
                     {
                         DVBufferPass = MPT_DVPASS_ECHO_BUFFER;
                         DVcmd = MPT_DV_READ_BUFFER_CAPACITY;
                         mpt_DVIoCmd ();
                     }
                 }
                 else
                 {
                     if ( DVdp->tlflags & MPT_RAID_VOLUME )
                         mpt_LUNComplete ( MPT_DV_SUCCESS, "RAID VOLUME Inq Success" );
                     else
                         mpt_LUNComplete ( MPT_DV_SUCCESS, "FIBRE Inq Success" );
                 }
            }
	}
    }
    else if ( ( CompCode == MPT_STAT_NOTEQ ) ||
             ( ( DVdp->tlflags & MPT_DV_DEVICE ) == 0 ) )
    {
#ifdef MPT_DEBUG
        flags = MPT_DV_SHOW_MSG | MPT_DV_FAIL;
#else
        flags = MPT_DV_FAIL;
#endif
        mpt_LUNComplete ( flags, "DVInq Failed: Selection Timeout" );
    }
    else if (CompCode == MPT_DID_TIMEOUT)
    {
        /*
         * Fall back and try to run DV again. Once we hit the
         * minimum speed, just return.
         */
        DVpass = 0;
        mpt_DVParms ( MPT_DV_SET_FALLBACK);
    }
    else
    {
        /*
         * Handle all other errors. If in debug mode,
         * print a message, other wise just set the flags.
         */
        if (CompCode == MPT_STAT_RETRY)
        {
            cmn_err (CE_NOTE, "DVInq: Retry on ha=%d id=%d lun=%d\n", DVpath, DVid, DVlun);
            mpt_DVIoCmd ();
        }
        else
        {
            /*
             * Catch all for error cases. 
             */
            /*
             * Fall back and try to run DV again. Once we hit the
             * minimum speed, just return.
             */
            DVpass = 3;
            mpt_DVParms ( MPT_DV_SET_FALLBACK);
        }
    }
}


/****************************************************************************
 *
 * Name:    mpt_DVDevicePage0Completed 
 * Summary: Handle the callback for the Read SCSI Device Page 0 command 
 *          via ReadConfigPage during the bus scan.   Note: ReadConfigPage
 *          allocates its own MF, so it needs to be freed here.
 *
 * Inputs:  Adapter structure, DVmhp, and command pointer, MFtabPtr
 * Returns: None.
 *
 ****************************************************************************/
void mpt_DVDevicePage0Completed ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    extern  char *g_ConfigBufPtr;
    ubit32 CompCode, vv;
    SCSIDevicePage0_t *dataBufPtr;
    uchar  ActualSyncFactor;

    CompCode = MFtabPtr->CompCode;

    /*
     * Command completed. If status was good,
     * save the data else just free the allocated structs and issue wakeup.
     */

    mpt_FreeCmdMFs (DVmhp, MFtabPtr->MFindex);
    if (CompCode == MPI_IOCSTATUS_SUCCESS)
    {
        dataBufPtr = (SCSIDevicePage0_t *)g_ConfigBufPtr;

        /*
         * Get Data
         */
        vv=dataBufPtr->NegotiatedParameters;
        SWAP_BYTES(vv);
        ActualSyncFactor = (vv & MPI_SCSIDEVPAGE0_NP_NEG_SYNC_PERIOD_MASK) >> 8;
        if ( ActualSyncFactor != DVdp->SyncFactor )
        {
            /*
             * Fall back and try to run DV again. Once we hit the
             * minimum speed, just return.
             */
            DVpass = 0;
            mpt_DVParms ( MPT_DV_SET_FALLBACK);
        }
        else
        {
            DVBufferPass = MPT_DVPASS_ECHO_BUFFER;
            DVcmd = MPT_DV_READ_BUFFER_CAPACITY;
            mpt_DVIoCmd ();
        }
    }
    else
    {
        /*
         * Handle all other errors. 
         */
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "DVDevicePage0 Failed" );
    }
}


/****************************************************************************
 *
 * Name:    mpt_DVQuiesceCompleted 
 * Summary: Handle the callback for the Quiesce command during the 
 *          bus scan.
 *
 * Inputs:  Adapter structure, DVmhp, and command pointer, MFtabPtr
 * Returns: None.
 *
 * Remarks: The scsi status and the completion code have been set by the ISR.  
 *          This routine must provide all necessary clean-up.   
 *
 ****************************************************************************/
void mpt_DVQuiesceCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    ubit32 CompCode, vv;
    char  flags;

    CompCode = MFtabPtr->CompCode;
    DVflags |= MPT_DV_DEVICE_QUIESCED;

    if (CompCode == MPI_RAID_ACTION_ASTATUS_SUCCESS )
    {
        mpt_DVConfigHeader ();
    }
    else
    {
        /*
         * Handle all other errors. 
         */
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "DVQuiesce Failed" );
    }
}

/****************************************************************************
 *
 * Name:    mpt_DVInqLunCompleted 
 * Summary: Handle the callback for the DVInquiry command for Luns > 0
 *
 * Inputs:  Adapter structure, DVmhp, and command pointer, MFtabPtr
 * Returns: None.
 *
 * Remarks: The scsi status and the completion code have been set by the ISR.  
 *          The Inquiry data has not been copied to the Dev_t structure.
 *          This routine must provide all necessary clean-up.   
 *
 ****************************************************************************/
void mpt_DVInqLunCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    SCSIIORequest_t *ioreq;
    ubit32 CompCode, vv;

    CompCode = MFtabPtr->CompCode;

    if (CompCode == MPT_STAT_GOOD)
    {
    	ioreq = (SCSIIORequest_t *)MFtabPtr->MFVirtAddr;
        vv = ioreq->DataLength;
        SWAP_BYTES ( vv );
	if ( vv < DVInqLength )
	    DVInqLength = vv;
        mpt_setDevParameters ( FirstDVBufPtr);
        mpt_LUNComplete ( MPT_DV_SUCCESS, "DVInqLun Success" );
    }
    else
        mpt_LUNComplete ( MPT_DV_FAIL, "DVInqLun Failed" );
}

/****************************************************************************
 *
 * Name:    mpt_DVReadBufferCapCompleted 
 * Summary: Handle the callback for the Read Buffer Capacity command during DV.
 *
 * Inputs:  Adapter structure, DVmhp, and command index
 * Returns: None.
 *
 * Remarks: DVpass == MPT_DVPASS_ECHO_BUFFER  -> testing w/ echo buffer.
 *          DVpass == MPT_DVPASS_DATA_BUFFER  -> using the regular buffer.
 *
 ****************************************************************************/
void mpt_DVReadBufferCapCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    ubit32 CompCode;
    ubit32   BufSize;

    CompCode = MFtabPtr->CompCode;

    if (CompCode == MPT_STAT_GOOD)
    {
        /*
         * If buffer capacity == 0
         *   If pass 0 reissue with descriptor mode
         *   If pass > 0 done
         * If buffer capacity > 0
         *    start write/read/compare buffer sequences.
         */
        BufSize = FirstDVBufPtr[1]<<16 | FirstDVBufPtr[2]<<8 | FirstDVBufPtr[3];

        if (BufSize == 0)
        {
            if ( DVBufferPass == MPT_DVPASS_ECHO_BUFFER)
            {
                DVBufferPass = MPT_DVPASS_DATA_BUFFER;
                mpt_DVIoCmd ();
            }
            else 
            {
#ifdef MPT_DEBUG
            flags = MPT_DV_SHOW_MSG | MPT_DV_SUCCESS;
#else
            flags = MPT_DV_SUCCESS;
#endif
                mpt_LUNComplete ( flags, "DVReadBufferCap Failed: No Buffers" );
            }
        }
        else 
        {   
            /*
             * Free buffer 
             */
//            kmem_free (FirstDVBufPtr, DVBufferCapacity);
            if ( BufSize > MPT_DV_MAX_BUF )
                DVBufferCapacity = MPT_DV_MAX_BUF;
            else
                DVBufferCapacity = BufSize;

            /*
             * Issue Reserve Command.
             */
            DVcmd = MPT_DV_RESERVE;
            mpt_DVIoCmd ();
        }
    }
    else if (CompCode == MPT_STAT_CKSTAT)
    {
        /* 
         * Check condition.
         */
        if (DVdp->SenseData[2] == MPT_SK_UNATTEN)
        {
            DVBufferPass = MPT_DVPASS_ECHO_BUFFER;  /* Retry ECHO_BUFFER */
            mpt_DVIoCmd ();
        }
        else if ( DVBufferPass == MPT_DVPASS_ECHO_BUFFER )
        {
            DVBufferPass = MPT_DVPASS_DATA_BUFFER;
            mpt_DVIoCmd ();
        }
        else 
        {
#ifdef MPT_DEBUG
            flags = MPT_DV_SHOW_MSG | MPT_DV_SUCCESS;
#else
            flags = MPT_DV_SUCCESS;
#endif
            mpt_LUNComplete ( flags, "DVReadBufferCap Failed: Check Condition" );
        }
    }
    else
    {
        if (CompCode == MPT_STAT_NOTEQ)
        {
            mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DVReadBufferCap Failed: Selection Timeout" );
        }
        else if (CompCode == MPT_STAT_RETRY)
        {
            cmn_err (CE_NOTE, "DVReadBufferCap: Retry on ha=%d id=%d lun=%d\n", DVpath, DVid, DVlun);
            mpt_DVIoCmd ();
        }
        else
        {
            /*
             * Fall back and try to run DV again. Once we hit the
             * minimum speed, just return.
             */
            DVBufferPass = MPT_DVPASS_DATA_BUFFER;
            mpt_DVParms ( MPT_DV_SET_FALLBACK);
        }
    }
}

/****************************************************************************
 *
 * Name:    mpt_DVReserveCompleted 
 * Summary: Handle the callback for the Reserve command during DV.
 *
 * Inputs:  Adapter structure, DVmhp, and command index
 * Returns: None.
 *
 * Remarks: DVpass == MPT_DVPASS_ECHO_BUFFER  -> testing w/ echo buffer.
 *          DVpass == MPT_DVPASS_DATA_BUFFER  -> using the regular buffer.
 *
 ****************************************************************************/
void mpt_DVReserveCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    ubit32 CompCode;
    ubit32   BufSize;

    CompCode = MFtabPtr->CompCode;

    if (CompCode == MPT_STAT_GOOD)
    {
        /*
         * Begin Write/Read/Compare loop. 
         * Fill write buffer with pattern (set by DVpattern) and issue write command
         */
        DVflags |= MPT_DV_DEVICE_RESERVED;
        mpt_DVFillBuffer ( DVdp->NegWidth );

        DVcmd = MPT_DV_WRITE_BUFFER;
        mpt_DVIoCmd ();
    }
    else if (CompCode == MPT_STAT_CKSTAT)
    {
        /* 
         * Check condition.
         */
#ifdef MPT_DEBUG
        flags = MPT_DV_SHOW_MSG | MPT_DV_SUCCESS;
#else
        flags = MPT_DV_SUCCESS;
#endif
        mpt_LUNComplete ( flags, "DVReserve Failed: Check Condition" );
    }
    else if (CompCode == MPT_DID_TIMEOUT)
    {
        /*
         * Treat this as a fatal error.
         */
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "DVReserve Failed: Timeout" );
    }
    else if (CompCode == MPT_DID_IOC_HARD_RESET)
    {
        /*
         * Fatal Error.
         */
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "DVReserve Failed: Hard Reset" );
    }
    else
    {
        if (CompCode == MPT_STAT_NOTEQ)
        {
            mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DVReserve: Selection Timeout" );
        }
        else if (CompCode == MPT_STAT_RETRY)
        {
            cmn_err (CE_NOTE, "DVReserve: Retry on ha=%d id=%d lun=%d\n", DVpath, DVid, DVlun);
            mpt_DVIoCmd ();
        }
        else
        {
            /*
             * Catch all for error cases. 
             */
#ifdef MPT_DEBUG
            flags = MPT_DV_SHOW_MSG | MPT_DV_FAIL;
#else
            flags = MPT_DV_FAIL;
#endif
            mpt_LUNComplete ( flags,"DVReserve" );
        }
    }
}

/****************************************************************************
 *
 * Name:    mpt_DVWriteBufferCompleted 
 * Summary: Handle the callback for the Write Buffer Capacity command during DV.
 *
 * Inputs:  Adapter structure, DVmhp, and command index
 * Returns: None.
 *
 * Remarks: DVpass == MPT_DVPASS_ECHO_BUFFER  -> testing w/ echo buffer.
 *          DVpass == MPT_DVPASS_DATA_BUFFER  -> using the regular buffer.
 *
 ****************************************************************************/
void mpt_DVWriteBufferCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    ubit32 CompCode;
    char flags;

    CompCode = MFtabPtr->CompCode;

    if (CompCode == MPT_STAT_GOOD)
    {
        /*
         * Issue Read Buffer
         */
        DVcmd = MPT_DV_READ_BUFFER;
        mpt_DVIoCmd ();
    }
    else if (CompCode == MPT_STAT_CKSTAT)
    {
        /* 
         * Check condition. A bus reset will cause a check condition,
         * and will have released the device. 
         * Issue a release. Since flags 0x80 or'd in, reserve and then
         * start the write buffer command again.  Any other cases result
         * in a failure. 
         */

        if (DVdp->SenseData[2] == MPT_SK_UNATTEN)
        {
            DVcmd = MPT_DV_RELEASE_THEN_RESERVE;
            mpt_DVIoCmd ();
        }
	else if ( DVdp->SenseData[2] == MPT_SK_ILLREQ )
	{
            DVBufferPass = MPT_DVPASS_DATA_BUFFER;
            DVcmd = MPT_DV_READ_BUFFER_CAPACITY;
            mpt_DVIoCmd ();
	}
        else
        {
            mpt_DVParms ( MPT_DV_SET_FALLBACK);
        }
    }
    else
    {
        if (CompCode == MPT_STAT_NOTEQ)
        {
            mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DV Failed: WriteBuffer Selection Timeout" );
        }
        else if (CompCode == MPT_STAT_RETRY)
        {
            cmn_err (CE_NOTE, "DVWriteBuffer: Retry on ha=%d id=%d lun=%d\n", DVpath, DVid, DVlun);
            mpt_DVIoCmd ();
        }
        else
        {
            /*
             * Fall back and try to run DV again. Once we hit the
             * minimum speed, just return.
             */
            mpt_DVParms ( MPT_DV_SET_FALLBACK);
        }
    }
}

/****************************************************************************
 *
 * Name:    mpt_DVReadBufferCompleted 
 * Summary: Handle the callback for the Read Buffer Capacity command during DV.
 *
 * Inputs:  Adapter structure, DVmhp, and command index
 * Returns: None.
 *
 * Remarks: DVpass == MPT_DVPASS_ECHO_BUFFER  -> testing w/ echo buffer.
 *          DVpass == MPT_DVPASS_DATA_BUFFER  -> using the regular buffer.
 *
 ****************************************************************************/
void mpt_DVReadBufferCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    ubit32 CompCode;
    char   flags;

    CompCode = MFtabPtr->CompCode;

    if ( CompCode == MPT_STAT_GOOD )
    {
        /*
         * Write/Read complete - do a compare.
         * If successful, get the next pattern
         * and call write again.
         */
        if (bcmp ((caddr_t)FirstDVBufPtr, (caddr_t)SecondDVBufPtr, DVBufferCapacity))
        {   
            if ( DVBufferPass == MPT_DVPASS_ECHO_BUFFER)
            {
                DVBufferPass = MPT_DVPASS_DATA_BUFFER;
                DVcmd = MPT_DV_READ_BUFFER_CAPACITY;
                mpt_DVIoCmd ();
            }
	    else
            {
                /*
                 * Miscompare -> do a fallback and start all over again.
                 */
                DVcmd = MPT_DV_WRITE_BUFFER;
                mpt_DVFillBuffer ( DVdp->NegWidth );
                mpt_DVParms ( MPT_DV_SET_FALLBACK);
            }
        }   
        else if (DVpattern < MPT_DV_MAX_PATTERN)
        {
            DVpattern++;
            mpt_DVFillBuffer ( DVdp->NegWidth );
            DVcmd = MPT_DV_WRITE_BUFFER;
            mpt_DVIoCmd ();
        }
        else 
        {
#ifdef MPT_DEBUG
            flags = MPT_DV_SHOW_MSG | MPT_DV_SUCCESS;
#else
            flags = MPT_DV_SUCCESS;
#endif
            mpt_LUNComplete ( flags, "DV Success" );
        }
    }
    else if (CompCode == MPT_STAT_CKSTAT)
    {
        /* 
         * Check condition. A bus reset will cause a check condition,
         * and will have released the device. 
         * Issue a release. Since flags 0x80 or'd in, reserve and then
         * start the write buffer command again.  Any other cases result
         * in a failure. 
         */

        if (DVdp->SenseData[2] == MPT_SK_UNATTEN)
        {
            DVcmd = MPT_DV_RELEASE_THEN_RESERVE;
            mpt_DVIoCmd ();
        }
	else if ( DVdp->SenseData[2] == MPT_SK_ILLREQ )
	{
            DVBufferPass = MPT_DVPASS_DATA_BUFFER;
            DVcmd = MPT_DV_READ_BUFFER_CAPACITY;
            mpt_DVIoCmd ();
	}
        else
        {
            DVcmd = MPT_DV_WRITE_BUFFER;
            mpt_DVFillBuffer ( DVdp->NegWidth );
            mpt_DVParms ( MPT_DV_SET_FALLBACK);
        }
    }
    else
    {
        if (CompCode == MPT_STAT_NOTEQ)
        {
            mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DVReadBuffer Failed: Selection Timeout" );
        }
        else if (CompCode == MPT_STAT_RETRY)
        {
            cmn_err (CE_NOTE, "DVReadBuffer: Retry on ha=%d id=%d lun=%d\n", DVpath, DVid, DVlun);
            mpt_DVIoCmd ();
        }
        else
        {
            /*
             * Fall back and try to run DV again. Once we hit the
             * minimum speed, just return.
             */
            DVcmd = MPT_DV_WRITE_BUFFER;
            mpt_DVFillBuffer ( DVdp->NegWidth );
            mpt_DVParms ( MPT_DV_SET_FALLBACK);
        }
    }
}

/****************************************************************************
 *
 * Name:    mpt_DVReleaseCompleted 
 * Summary: Handle the callback for the Release command during DV.
 *
 * Inputs:  Adapter structure, DVmhp, and command index
 * Returns: None.
 *
 * Remarks: DVpass == MPT_DVPASS_ECHO_BUFFER  -> testing w/ echo buffer.
 *          DVpass == MPT_DVPASS_DATA_BUFFER  -> using the regular buffer.
 *
 ****************************************************************************/
void mpt_DVReleaseCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    ubit32 CompCode;
    char   flags;

    CompCode = MFtabPtr->CompCode;
    DVflags &= ~MPT_DV_DEVICE_RESERVED;

    if (CompCode == MPT_STAT_GOOD)
    {

        if (DVcmd & 0x80)
        {
            /*
             * The write/read/compare loop was interrupted by a bus reset.
             * Issue the reserve and continue.
             */
            DVcmd = MPT_DV_RESERVE;
            mpt_DVIoCmd ();
        }
        else
        {
            /*
             * We are done.
             */
#ifdef MPT_DEBUG
            flags = MPT_DV_SHOW_MSG | MPT_DV_SUCCESS;
#else
            flags = MPT_DV_SUCCESS;
#endif
            mpt_LUNComplete ( flags, "DV Success" );
        }
    }
    else if (CompCode == MPT_STAT_CKSTAT)
    {
        /* 
         * Check condition.
         */
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_SUCCESS, "DVRelease Failed: Check Condition" );
    }
    else if (CompCode == MPT_DID_TIMEOUT)
    {
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "DVRelease Failed: Timeout" );
    }
    else if (CompCode == MPT_DID_IOC_HARD_RESET)
    {
        /*
         * Fatal Error.
         */
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL, "DVRelease Failed: Hard Reset" );
    }
    else
    {
        if (CompCode == MPT_STAT_NOTEQ)
        {
            mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DVRelease Failed: Selection Timeout" );
        }
        else if (CompCode == MPT_STAT_RETRY)
        {
            cmn_err (CE_NOTE, "DVRelease: Retry on ha=%d id=%d lun=%d\n", DVpath, DVid, DVlun);
            mpt_DVIoCmd ();
        }
        else
        {
            mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DVRelease Failed" );
        }
    }
}

/****************************************************************************
 *
 * Name:    mpt_DVEnableCompleted 
 * Summary: Handle the callback for the Enable command during DV.
 *
 * Inputs:  Adapter structure, DVmhp, and command index
 * Returns: None.
 *
 * Remarks: DVpass == MPT_DVPASS_ECHO_BUFFER  -> testing w/ echo buffer.
 *          DVpass == MPT_DVPASS_DATA_BUFFER  -> using the regular buffer.
 *
 ****************************************************************************/
void mpt_DVEnableCompleted ( pHBA_t DVmhp, TableStruct_t *MFtabPtr )
{
    ubit32 CompCode;
    char   flags;

    CompCode = MFtabPtr->CompCode;

    DVflags &= ~MPT_DV_DEVICE_QUIESCED;

    if (CompCode == MPI_RAID_ACTION_ASTATUS_SUCCESS )
    {
        /*
         * We are done.
         */
#ifdef MPT_DEBUG
        flags = MPT_DV_SHOW_MSG | MPT_DV_DONE;
#else
        flags = MPT_DV_DONE;
#endif
        mpt_LUNComplete ( flags, "DV Success" );
    }
    else
    {
        mpt_LUNComplete ( MPT_DV_SHOW_MSG | MPT_DV_FAIL,"DVEnable Failed" );
    }
    DVmhp->HBA_Chars &= ~MPT_DV_PHYSDISK_IN_PROGRESS;  /* PhysDisk DV is Completed */
    mpt_StartQueuedCommands(DVmhp);
}


/****************************************************************************
 *
 * Name:    mpt_DVFillBuffer
 * Summary: Populate the FirstPhysAddr (FirstDVBufPtr) with a pattern specified
 *          by the global variable DVpattern.
 *
 * Inputs:  None.
 * Returns: None.
 *
 * Remarks: 
 *
 ****************************************************************************/
void mpt_DVFillBuffer ( uchar width )
{
    uchar *ptr = FirstDVBufPtr;
    int ii;
    char byte;
    short val;

    switch (DVpattern)
    {
        case 0:
            
            if (width)
            {
                /*
                 * Pattern:  0000 FFFF 0000 FFFF
                 */
                for (ii=0; ii < DVBufferCapacity; ii++, ptr++)
                {
                   if (ii & 0x02)
                   {
                       *ptr = 0xFF;
                   }
		   else
			*ptr = 0x00;
                }
            }
            else
            {
                /*
                 * Pattern:  00 FF 00 FF
                 */
                for (ii=0; ii < DVBufferCapacity; ii++, ptr++)
                {
                   if (ii & 0x01)
                   {
                       *ptr = 0xFF;
                   }
		   else
			*ptr = 0x00;
                }
            }
            break;

        case 1:
            
            if (width)
            {
                /*
                 * Pattern:  5555 AAAA 5555 AAAA 5555
                 */
                for (ii=0; ii < DVBufferCapacity; ii++, ptr++)
                {
                   if (ii & 0x02)
                   {
                       *ptr = 0xAA;
                   }
		   else
			*ptr = 0x55;
                }
            }
            else
            {
                /*
                 * Pattern:  55 AA 55 AA 55
                 */
                for (ii=0; ii < DVBufferCapacity; ii++, ptr++)
                {
                   if (ii & 0x01)
                   {
                       *ptr = 0xAA;
                   }
		   else
			*ptr = 0x55;
                }
            }
            break;

        case 2:
            
            if (width)
            {
                /*
                 * Wide Pattern:  0000 0001 0002 0003 0004  ... FFFE FFFF 0000 0001..
                 */
                for (ii=0; ii < DVBufferCapacity/2; ii++)
                {
                    *ptr = (char) ((ii & 0xFF00) >> 8);
                    ptr++;
                    *ptr = (char) (ii & 0xFF);
                    ptr++;
                }
            }
            else
            {
                /*
                 * Narrow Pattern:  00 01 02 03 04 05 ... FE FF 00 01..
                 */
                for (ii=0; ii < DVBufferCapacity; ii++, ptr++)
                {
                   *ptr = (char) ii;
                }
            }

            break;

        case 3:
            
            if (width)
            {
                /* 
                 * Wide Pattern:  FFFE 0001 FFFD 0002 ...  4000 DFFF 8000 EFFF
                 */
                byte = 0;
                for (ii=0; ii < DVBufferCapacity/2; ii++)
                {
                    /*
                     * Create the base pattern
                     */
                    val = (1 << byte);
                    /*
                     * every 64 (0x40) bytes flip the pattern
                     * since we fill 2 bytes / iteration, test for ii = 0x20
                     */
                    if (ii & 0x20) 
                    {
                        val = ~(val);
                    }

                    if (ii & 0x01)
                    {
                        *ptr = (char)( (val & 0xFF00) >> 8);
                        ptr++;
                        *ptr = (char)(val & 0xFF);

                        byte++;
                        byte &= 0x0F;
                    }
                    else
                    {
                        val = ~val;
                        *ptr = (char)( (val & 0xFF00) >> 8);
                        ptr++;
                        *ptr = (char)(val & 0xFF);
                    }

                    ptr++;
                }
            }
            else
            {
                /* 
                 * Narrow Pattern:  FE 01 FD 02 FB 04 .. 7F 80 01 FE 02 FD ...  80 7F
                 */
                byte = 0;
                for (ii=0; ii < DVBufferCapacity; ii++, ptr++)
                {
                    /* 
                     * Base pattern - first 32 bytes 
                     */
                    if (ii & 0x01)
                    {
                        *ptr = (1 << byte);
                        byte++;
                        byte &= 0x07;
                    }
                    else
                    {
                        *ptr = (char) (~(1 << byte));
                    }

                    /*
                     * Flip the pattern every 32 bytes
                     */
                    if (ii & 0x20)
                    {
                        *ptr = ~(*ptr);
                    }
                }
            }
            break;
    }
}

/****************************************************************************
 *
 * Name:    mpt_setDevParameters 
 * Summary: Set the Device parameters based on the inquiry data
 *      
 *
 * Inputs:  HBA ptr and a device ptr
 * Returns: MPT_FAILURE (0) or MPT_SUCCESS (1)
 *
 * Remarks:
 * 
 ****************************************************************************/
void  mpt_setDevParameters ( uchar *from)
{
    extern ulong_t mpt_TapeCommandTimeout;
    uchar version, type;
 
    if ( DVInqLength && ( ( from[0] & 0xE0 ) == 0 ) )
    {
        /*
         * This is a valid LUN, update Dev_t structure
         */
      
        DVdp->luns |= ( 1<<DVlun );
        DVflags |= MPT_DV_INQ_DONE;

        if ((DVdp->tlflags & MPT_INQUIRY_DATA_VALID) == 0)
        {
            /*
             * Set the Dev_t parameters, if this is new inquiry data
             */
            DVdp->tlflags |= MPT_INQUIRY_DATA_VALID;

	    DVdp->Type = type = from[0];
            if ( type == 0x03 || type == 0x0d )
                DVdp->tlflags &= ~MPT_DV_DEVICE;
            else if ( type == TAPE && DVdp->Timeout < mpt_TapeCommandTimeout )
                DVdp->Timeout = mpt_TapeCommandTimeout;
                
            version = from[2] & 0x07;
            DVdp->SCSI_Version = version;

            if (version >= 2)
            {
                if ( (from[7] & 0x60) == 0 )
                {
                    DVdp->tlflags |= MPT_NO_WDTR;
                    DVdp->Width = 0;
                    if ( DVdp->SyncFactor < MPT_ULTRA2 )
                        DVdp->SyncFactor = MPT_ULTRA2;
                }
                else
                {
        	    if ( ( devNvram != MPT_NVRAM_UNKNOWN ) && ( devNvram & MPT_NVRAM_WIDE_DISABLE ) )
                    {
                        DVdp->tlflags |= MPT_NO_WDTR;
                        DVdp->Width = 0;
                        if ( DVdp->SyncFactor < MPT_ULTRA2 )
                            DVdp->SyncFactor = MPT_ULTRA2;
                    }
                    else
                    {
                        DVdp->tlflags &= ~MPT_NO_WDTR;
                        DVdp->Width = MPT_WIDE;
                    }
                }

                if ( (from[7] & 0x10) == 0 )
                {
                    DVdp->tlflags |= MPT_NO_SDTR;
                    DVdp->Offset = 0;
                }

                if ( (from[7] & 0x02) == 0 )
        	    DVdp->Control |= MPI_SCSIIO_CONTROL_UNTAGGED;
                else
                {
        	    if ( ( devNvram != MPT_NVRAM_UNKNOWN ) && ( (devNvram & MPT_NVRAM_TAG_QUEUE_ENABLE) == 0 ) )
        	        DVdp->Control |= MPI_SCSIIO_CONTROL_UNTAGGED;
                }
 
                if ( DVmhp->IOCPage4Ptr &&
                     from[0] == 0x03 && from[44] == 'S' &&
                     from[45] == 'A' && from[46] == 'F' &&
                     from[47] == '-' && from[48] == 'T' &&
                     from[49] == 'E' )
                {
                    mpt_DVWritePage4 ();
                }
            }
            else
            {
                /*
                 * ANSI I supports only narrow devices.
                 */
                DVdp->tlflags |= MPT_NO_WDTR;
                DVdp->Width = 0;
                if ( DVdp->SyncFactor < MPT_ULTRA2 )
                     DVdp->SyncFactor = MPT_ULTRA2;
        	DVdp->Control |= MPI_SCSIIO_CONTROL_UNTAGGED;
            }
        } /* end of type valid */
    }  /* End of type & 0xe0 = 0 */
}

/****************************************************************************
 *
 * Name:    mpt_DVWritePage4
 * Summary: Issue the Message to write the IOCPage 4
 *
 * Inputs:  device pointer
 * Returns: None.
 *
 * Remarks: Construct a Config command for the given inputs.
 *          Add to the HBA driver_q.  Issue the command and wait for interrupt.
 *
 *          call back: DVWritePageCompleted 
 *
 *          Hba lock held upon entry and exit.
 *
 *          If DVpass = -1, do a write commit.
 *
 ****************************************************************************/
void mpt_DVWritePage4 (void)
{
    TableStruct_t *MFtabPtr;
    Config_t    *Request;
    SGESimple32_t *mpisge;
    IOCPage4_t *IOCPage4Ptr;
    ubit32      MFindex, vv, ii;
    
    if ( MFtabPtr = mpt_GetFreeMF (DVmhp) )
    {
        MFindex = MFtabPtr->MFindex;
        MFtabPtr->dp = DVdp;
    }
    else
    {
        cmn_err ( CE_WARN, "IOCPage4 Write Failed: No MF available" );
        return;
    }

        /*
         * Populate the TableStruct_t
         * Use a generic LUN of 0 since DVdp is defined.
         */
        MFtabPtr->CallBackPtr = NULL;
        MFtabPtr->lun = 0;
        MFtabPtr->Flags = 0;

        /*
         * Populate the Message Frame 
         */
        Request = (Config_t *) MFtabPtr->MFVirtAddr;
        Request->Function = MPI_FUNCTION_CONFIG;
        Request->ChainOffset = 0;
        Request->Action = MPI_CONFIG_ACTION_PAGE_WRITE_CURRENT;
        Request->MsgFlags = 0;  

#ifdef _HPUX_SOURCE
        Request->MsgContext = MFindex<<8;
#else
        Request->MsgContext = MFindex;
#endif  

        IOCPage4Ptr = DVmhp->IOCPage4Ptr;
        ii = IOCPage4Ptr->ActiveSEP++;
        IOCPage4Ptr->SEP[ii].SEPTargetID = DVid;
        IOCPage4Ptr->SEP[ii].SEPBus = DVbus;
        Request->Header = IOCPage4Ptr->Header;
        vv =  ( DVbus << 8 ) | DVid;
	SWAP_BYTES (vv);
        Request->PageAddress = vv;

        mpisge = (SGESimple32_t *) &Request->PageBufferSGE;

        vv = ( MPT_SGE_FLAGS_SIMPLE_ELEMENT | 
	       MPT_SGE_FLAGS_32_BIT_ADDRESSING |
               MPT_SGE_FLAGS_HOST_TO_IOC |
	       MPT_SGE_FLAGS_LAST_ELEMENT |
	       MPT_SGE_FLAGS_END_OF_BUFFER |
               MPT_SGE_FLAGS_END_OF_LIST ) |
               (IOCPage4Ptr->Header.PageLength + ii) * 4;
        SWAP_BYTES(vv);
        mpisge->FlagsLength = vv;

        vv = DVmhp->IOCPage4PhysAddr;
        SWAP_BYTES(vv);
        mpisge->Address = vv;
        DVMFtabPtr->CompCode = MPT_STAT_GOOD;

        mpt_StartNewCommand (DVmhp, MFtabPtr);
}
